package discords;

import dist.DistEuclidean;

import java.util.Vector;
import java.io.FileOutputStream;

import utils.IOUtils;
import utils.Utils;
import datamgr.DataReader;

/**
 * Created by IntelliJ IDEA.
 * User: dragomir
 * Date: Oct 4, 2007
 * Time: 5:52:46 PM
 * To change this template use File | Settings | File Templates.
 */
public class Discords {
    public static char delim = '\t';
    public static boolean normalize = true;

    public void findDiscords(String dbDir, int numDbPages, double threshold, int k) throws Exception {
        Vector vMM = new Vector(); //the main memory
        Vector indxPtr = new Vector(); //indices of the elements in the main memory
        Vector vVolume = new Vector(); //memory volume after reading each time series
        MainMemory mm = new MainMemory();
        long startTime = System.currentTimeMillis();

        System.out.println("=====================================================================================");
        //Building the main memory with all DB pages
        System.out.println("Phase1: Building main memory.........................");
        int offset = 0;
        long iotime = 0;
        long iostarttime = 0;
        long ioendtime = 0;
        FileOutputStream fos = new FileOutputStream(dbDir + "javaversion_memory.txt");
        for (int i = 0; i < numDbPages; i++) {
            iostarttime = System.currentTimeMillis();
            Vector dbPage = loadPage(dbDir, i, numDbPages);
            ioendtime = System.currentTimeMillis();
            iotime += (ioendtime - iostarttime);
            mm.buildMainMemory(dbPage, threshold, vMM, vVolume, indxPtr, offset);
            offset += dbPage.size();
            fos.write(("" + vMM.size() + "\n").getBytes());
        }
        long timeDuration = System.currentTimeMillis() - startTime;
        String strDuration = "Phase1 finished. Time: " + (timeDuration / 1000) + " sec.";
        fos.write((strDuration + "\n").getBytes());
        System.out.println(strDuration);
        System.out.println("Number of discords = " + vMM.size());
        System.out.println("------------------------------------------------------");

        //Refining the discord candidates
        System.out.println("Phase2: Discord refinement............................");
        offset = 0;
        Vector nnDists = new Vector();
        startTime = System.currentTimeMillis();
        for (int i = 0; i < vMM.size(); i++) {
            nnDists.add(new Double(Double.MAX_VALUE));
        }
        for (int i = 0; i < numDbPages; i++) {
            iostarttime = System.currentTimeMillis();
            Vector dbPage = loadPage(dbDir, i, numDbPages);
            ioendtime = System.currentTimeMillis();
            iotime += (ioendtime - iostarttime);
            findDiscords(dbPage, vMM, nnDists, indxPtr, vVolume, threshold, offset);
            offset += dbPage.size();
            fos.write(("" + vMM.size() + "\n").getBytes());
        }
        long timeDuration1 = System.currentTimeMillis() - startTime;
        long totalTime = timeDuration + timeDuration1;
        String strDuration1 = "Phase2 finished. Time: " + (timeDuration1 / 1000) + " sec.";
        fos.write((strDuration1 + "\n").getBytes());
        System.out.println(strDuration1);
        String strtotalTime = "Phase1 & Phase2 time: " + (totalTime / 1000) + " sec.";
        fos.write((strtotalTime + "\n").getBytes());
        System.out.println(strtotalTime);
        String strIOTime = "IOTime = " + (iotime/1000) + " sec.";
        fos.write((strIOTime + "\n").getBytes());
        System.out.println(strIOTime);
        System.out.println("Number of discords = " + vMM.size());
        fos.close();

        //Output the top k discords
        System.out.println("------------------------------------------------------");
        System.out.println("Logging discords............................");
        startTime = System.currentTimeMillis();
        logDiscords(dbDir, vMM, nnDists, indxPtr, vVolume, k);
        timeDuration = System.currentTimeMillis() - startTime;
        System.out.println("Logging finished. Time: " + (timeDuration / 1000) + " sec.");
    }

    /**
     * Finds the top k discords based on the threshold
     */
    private void findDiscords(Vector DB, Vector MM, Vector nnDists, Vector indxPtr, Vector vVolume, double threshold, int offset) {
        DistEuclidean dist = new DistEuclidean(true, false);
        for (int i = 0; i < MM.size(); ) {
            boolean toRemove = false;
            int elIndx = ((Integer) indxPtr.get(i)).intValue();
            double currMinNNDist = ((Double) nnDists.get(i)).doubleValue();
            for (int j = 0; j < DB.size(); j++) {
                int pos = offset + j;
                if (pos == elIndx) {
                    continue;
                }
                double d = dist.dist((double[]) MM.get(i), (double[]) DB.get(j), currMinNNDist);
                if (d < threshold) {
                    toRemove = true;
                    break;
                }
                if (d < currMinNNDist) {
                    nnDists.set(i, new Double(d));
                    currMinNNDist = d;
                }
            }
            if (toRemove) {
                MM.remove(i);
                nnDists.remove(i);
                indxPtr.remove(i);
            } else {
                i++;
            }
        }
        vVolume.add(new Integer(MM.size()));
    }

    private Vector loadPage(String dir, int pageNum, int db_pages) throws Exception {
//        String fileName = dir + "TS";
//        int num_digits_allpages = (int) Math.ceil(Math.log(db_pages-1)/Math.log(10));
//        if ((double) num_digits_allpages == Math.log(db_pages-1)/Math.log(10)) {
//            num_digits_allpages++;  //when the number of pages is e.g. 101 wee need 3 digits
//        }
//        int num_digits_tmppage = 1;
//        if (pageNum != 0) {
//            num_digits_tmppage = (int) Math.ceil(Math.log(pageNum)/Math.log(10));
//            if ((double) num_digits_tmppage == Math.log(pageNum)/Math.log(10)) {
//                num_digits_tmppage++; //if e.g. pageNum == 100 we need 3 digits
//            }
//        }
//        int zero_nums = num_digits_allpages - num_digits_tmppage;
//        String suff = "";
//        int counter = 0;
//        while (counter < zero_nums) {
//            suff += "0";
//            counter *= 10;
//        };
//        fileName += (suff + pageNum + ".txt");
        String fileName = dir + "TS" + pageNum + ".txt";
        System.out.println("Loading page: " + fileName);
        return loadPage(fileName);
    }

    private Vector loadPage(String fileName) throws Exception {
        Vector dbPage = new Vector();
        DataReader reader = new DataReader(fileName, delim);
        String[] tokens;
        while ((tokens = reader.readLine()) != null) {
            double[] arr = new double[tokens.length];
            for (int i = 0; i < tokens.length; i++) {
                arr[i] = Double.parseDouble(tokens[i]);
            }
            if (normalize) {
                arr = Utils.znormalize(arr, 0, arr.length);
            }
            dbPage.add(arr);
        }
        reader.close();
        return dbPage;
    }

    private void logDiscords(String dir, Vector vMM, Vector nnDists, Vector indxPtr, Vector vVolume, int k) {
        try {
            sortDiscords(vMM, nnDists, indxPtr);
            FileOutputStream fos = new FileOutputStream(dir + "javaversion_discords");
            FileOutputStream fos1 = new FileOutputStream(dir + "javaversion_indices_dist_volume");
            int min = (k < vMM.size()) ? k : vMM.size();
            fos1.write("The first index is 0\n".getBytes());
            for (int i = 0; i < min; i++) {
                String line = "";
                double[] ts = (double[]) vMM.get(i);
                for (int j = 0; j < ts.length; j++) {
                    line += (IOUtils.df4.format(ts[j]) + "\t");
                }
                line = line.trim();
                line += "\n";
                fos.write(line.getBytes());
                String line1 = ("Index= " + ((Integer) indxPtr.get(i)).intValue() + "\tnnDist= " + (IOUtils.df4.format(((Double) nnDists.get(i)).doubleValue())));
                line1 += "\n";
                fos1.write(line1.getBytes());
            }
            fos1.write(("Memory volume during the two phases:\n").getBytes());
            for (int i = 0; i < vVolume.size(); i++) {
                String line = ("" + ((Integer) vVolume.get(i)).intValue() + "\n");
                fos1.write(line.getBytes());
            }
            fos1.close();
            fos.close();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * Sort the discords according to their nndists in descending order
     */
    private void sortDiscords(Vector vMM, Vector nnDists, Vector indxPtr) {
        for (int i = 0; i < vMM.size(); i++) {
            for (int j = i+1; j < vMM.size(); j++) {
                double tmpDist = ((Double) nnDists.get(i)).doubleValue();
                double tmpDist1 = ((Double) nnDists.get(j)).doubleValue();
                if (tmpDist < tmpDist1) {
                    nnDists.set(i, new Double(tmpDist1));
                    nnDists.set(j, new Double(tmpDist));

                    double[] tmpTs = (double[]) vMM.get(i);
                    vMM.set(i, vMM.get(j));
                    vMM.set(j, tmpTs);

                    Integer tmpPtr = (Integer) indxPtr.get(i);
                    indxPtr.set(i, indxPtr.get(j));
                    indxPtr.set(j, tmpPtr);
                }
            }
        }
    }

    public static void main(String[] args) throws Exception {
//        String dir = "./data/lightcurves/";
//        int numDbPages = 10;
//        double threshold = 10;
//        int numDiscords = 10;
////        String dir = "F:/data/randomwalk_100mill_size512/";
////        int numDbPages = 1;
////        double threshold = 21.45;
////        int numDiscords = 10000;

        int numDiscords = Integer.parseInt(args[0]);
        double threshold = Double.parseDouble(args[1]);
        String dir = args[2];
        int numDbPages = Integer.parseInt(args[3]);

        normalize = false;

        Discords disc = new Discords();
        disc.findDiscords(dir, numDbPages, threshold, numDiscords);
    }
}
