package utils;

import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: dragomir yankov
 * Date: Jun 20, 2005
 * Time: 11:35:26 AM
 * To change this template use File | Settings | File Templates.
 */
public class Utils {
    public static Random rnd = new Random(System.currentTimeMillis());

    /**
     * normalization with zero mean, variance 1:
     * x_i = (x_i - mean(x_i))/std(x_i), where
     * mean(x_i) = Sum(x_i)/n
     * std(x_i) = Sqrt(x_i - mean(x_i))/(n-1)
     */
    public static double[] znormalize(double[] data, int start, int end) {
        int n = end - start;
        double[] normdata = new double[n];
        double mean = mean(data, start, end);
        double std = std(data, start, end, mean, true);
        for (int i = 0; i < n; i++) {
            if (std == 0) {
                normdata[i] = 0;
            } else {
                normdata[i] = (data[start + i] - mean) / std;
            }
        }
        return normdata;
    }

    /**
     * reverses znormalization using: x_i = std*x_inormalized + mean
     */
    public static double[] unnormalize(double[] normdata, double mean, double std, int start, int end) {
        int n = end - start;
        double[] data = new double[n];
        for (int i = 0; i < n; i++) {
            data[i] = std*normdata[start + i] + mean;
        }
        return data;
    }

    public static double[] normalize(double[] data, double mean, double std, int start, int end) {
        int n = end - start;
        double[] normdata = new double[n];
        for (int i = 0; i < n; i++) {
            if (std == 0) {
                normdata[i] = 0;
            } else {
                normdata[i] = (data[start + i] - mean) / std;
            }
        }
        return normdata;
    }

    public static double[] znormalize_sparse(double[][] col_val, int start_ind, int end_ind) {
        double[] full = full(col_val, start_ind, end_ind);
        return znormalize(full, 0, full.length);
    }

    public static double mean(double[] data, int start, int end) {
        double sum = 0;
        for (int i = start; i < end; i++) {
            sum += data[i];
        }
        return sum / (end - start);
    }

    public static double mean(double[] data, double start, double end) {
        int intStart = (int) Math.ceil(start);
        int intEnd = (int) Math.floor(end);

        double sum = 0;
        double count = 0;
        if (intStart > start) {
            double weight = intStart - start;
            sum = data[intStart - 1] * weight;
            count = weight;
        }
        for (int i = intStart; i < intEnd; i++) {
            sum += data[i];
        }
        count += (intEnd - intStart);
        if (intEnd < end && intEnd > -1) {
            if (intEnd == 316) {
             System.out.println("Dump the dump");
            }
            double weight = end - intEnd;
            sum += data[intEnd]*weight;
            count += weight;
        }
        return sum / count;
    }

    public static double std(double[] data, int start, int end) {
        return std(data, start, end, 0, false);
    }

    public static double std(double[] data, int start, int end, double mean, boolean hasMean) {
        if (data.length <= 1 || (end - start <= 0)) {
            return 0;
        }

        int n = end - start;
        if (n == 1) {
            //return data[start];????????
            return 0;
        }

        if (!hasMean) {
            mean = mean(data, start, end);
        }
        double sum = 0;
        double nextel = 0;
        for (int i = start; i < end; i++) {
            nextel = (data[i] - mean);
            sum += nextel * nextel;
        }
        sum /= (n - 1);
        return Math.sqrt(sum);
    }

    public static double weighted_mean(double[] arr, double[] weights) {
        double mean = 0;
        for (int i = 0; i < arr.length; i++) {
            mean += arr[i] * weights[i];
        }
        mean /= sum(weights, 0, weights.length);
        return mean;
    }

    public static double weighted_std(double[] arr, double[] weights) {
        if (arr.length <= 1) {
            return 0;
        }
        double s_wi = sum(weights, 0, weights.length);
        double ssq_wi = sum_squares(weights);
        double mean = weighted_mean(arr, weights);

        double std = 0;
        for (int i = 0; i < arr.length; i++) {
            std += weights[i] * ((arr[i] - mean) * (arr[i] - mean));
        }
        std *= (s_wi / ((s_wi*s_wi) - ssq_wi));
        return Math.sqrt(std);
    }

    public static double sum(double[] arr, int start, int end) {
        double sum = 0;
        for (int i = start; i < end; i++) {
            sum += arr[i];
        }
        return sum;
    }

    public static double sum_squares(double[] arr) {
        double sum = 0;
        for (int i = 0; i < arr.length; i++) {
            sum += (arr[i] * arr[i]);
        }
        return sum;
    }

    /**
     * convert a sparse vector to full size vector
     * @param col_val
     * @param start_ind - start position in sparse vector
     * @param end_ind - end position in sparse vector
     * @return
     */
    public static double[] full(double[][] col_val, int start_ind, int end_ind) {
        if (end_ind < start_ind) {
            return null;
        }
        double[] fullvect = new double[end_ind + 1];
        for (int i = 0; i < col_val.length; i++) {
            if (col_val[i][0] >= start_ind && col_val[i][0] <= end_ind) {
                fullvect[(int) col_val[i][0]] = col_val[i][1];
            }
        }
        double[] fullvect_cut = new double[end_ind - start_ind + 1];
        System.arraycopy(fullvect, start_ind, fullvect_cut, 0, fullvect_cut.length);
        return fullvect_cut;
    }

    /**
     * Selects a random subset with a predifined capacity from the provided set
     * @param set
     * @param capacity
     * @return
     */
    public static HashSet rndSubset(HashSet set, int capacity) {
        Vector pos = new Vector();
        pos.addAll(set);
        HashSet rndSet = new HashSet(capacity);

        for (int i = 0; i < capacity; i++) {
            int nextRnd = rnd.nextInt(pos.size());
            rndSet.add(pos.get(nextRnd));
            pos.removeElementAt(nextRnd);
        }
        return rndSet;
    }

    /**
     * Splits a set into nonoverlaping random subsets
     * @param set
     * @param subsets
     * @return
     */
    public static HashSet[] split(HashSet set, int subsets) {
        int capacity = set.size() / subsets;
        HashSet[] hsubsets = new HashSet[subsets];

        for (int i = 0; i < subsets; i++) {
            hsubsets[i] = rndSubset(set, capacity);
        }
        return hsubsets;
    }

    public static double[] toDoubleArray(int[] v) {
        double[] arr = new double[v.length];
        for (int i = 0; i < v.length; i++) {
            arr[i] = v[i];
        }
        return arr;
    }

    public static double[] toDoubleArray(Vector v) {
        double[] arr = new double[v.size()];
        for (int i = 0; i < v.size(); i++) {
            arr[i] = ((Double) v.get(i)).doubleValue();
        }
        return arr;
    }

    public static int[] toIntArray(Vector v) {
        int[] arr = new int[v.size()];
        for (int i = 0; i < v.size(); i++) {
            arr[i] = ((Integer) v.get(i)).intValue();
        }
        return arr;
    }

    public static int[][] toIntIntArray(Vector v) {
        int[][] arr = new int[v.size()][];
        for (int i = 0; i < v.size(); i++) {
            arr[i] = (int[]) v.get(i);
        }
        return arr;
    }

    public static double[][] toDblDblArray(Vector v) {
        double[][] arr = new double[v.size()][];
        for (int i = 0; i < v.size(); i++) {
            arr[i] = (double[]) v.get(i);
        }
        return arr;
    }

    public static double[][][] toDblDblDblArray(Vector v) {
        double[][][] arr = new double[v.size()][][];
        for (int i = 0; i < v.size(); i++) {
            arr[i] = (double[][]) v.get(i);
        }
        return arr;
    }

    public static String[] toStrArray(Vector v) {
        String[] arr = new String[v.size()];
        for (int i = 0; i < v.size(); i++) {
            arr[i] = (String) v.get(i);
        }
        return arr;
    }

    public static double[][] transpose(double[][] M) {
        double[][] NM = new double[M[0].length][];
        for (int i = 0; i < NM.length; i++) {
            double[] row = new double[M.length];
            for (int j = 0; j < row.length; j++) {
                row[j] = M[j][i];
            }
            NM[i] = row;
        }
        return NM;
    }

    public static double[][] transpose(double[] M) {
        double[][] NM = new double[M.length][];
        for (int i = 0; i < NM.length; i++) {
            NM[i] = new double[1];
            NM[i][0] = M[i];
        }
        return NM;
    }

    public static int[] tableToIntArray(Hashtable table) {
        int[] arr = new int[table.size()];
        Enumeration en = table.keys();
        for (int i = 0; en.hasMoreElements(); i++) {
            arr[i] = ((Integer) en.nextElement()).intValue();
        }
        return arr;
    }

    public static int[] setToIntArray(HashSet set) {
        int[] arr = new int[set.size()];
        Iterator en = set.iterator();
        for (int i = 0; en.hasNext(); i++) {
            arr[i] = ((Integer) en.next()).intValue();
        }
        return arr;
    }

    public static Hashtable intArrToTable(int[] arr) {
        Hashtable table = new Hashtable();
        for (int i = 0; i < arr.length; i++) {
            table.put(new Integer(arr[i]), "1");
        }
        return table;
    }

    /**
     * @param array should be sorted
     */
    public static double[] quantile(double[] array, double quantile) {
        int newLength = (int) ((double)quantile*array.length);
        double[] newArr = new double[newLength];
        System.arraycopy(array, 0, newArr, 0, newLength);
        return newArr;
    }

    public static double min(double[] array, int begin, int end) {
        double min = array[begin];
        for (int i = begin+1; i < end; i++) {
            if (min > array[i]) {
                min = array[i];
            }
        }
        return min;
    }

    public static double max(double[] array, int begin, int end) {
        double max = array[begin];
        for (int i = begin+1; i < end; i++) {
            if (max < array[i]) {
                max = array[i];
            }
        }
        return max;
    }

    public static boolean startsWithDigit(String s) {
        String firstSymbol = s.substring(0, 1);
        try {
            int i = Integer.parseInt(firstSymbol);
            return true;
        } catch (Exception e) {
            return false;
        }
    }
}